import type { EventDetailMap } from '@genesislcap/foundation-events';
import { BindingObserver, Constructable, ExecutionContext } from '@microsoft/fast-element';
import { Key } from '@microsoft/fast-foundation';
import { Observable as RXObservable } from 'rxjs';
import { ErrorMap } from './errorMap';
/**
 * @typeParam TReturn - The return type.
 *
 * @public
 */
export type SubscriberChangeCallback<TReturn = any> = (value: TReturn, args: BindingObserver) => void;
/**
 * Store interface.
 *
 * @public
 * @remarks
 * The public interface available on the injected store fragment.
 */
export interface Store {
    /**
     * The name of the store fragment.
     * @returns string
     *
     * @virtual
     * @privateRemarks
     * This is used for debugging only, and is overridden during registerStore.
     */
    readonly name?: string;
    /**
     * Contains any errors the store may have, see {@link ErrorMap}.
     *
     * @example
     * Errors can be looked up by event key:
     * ```ts
     * ${when(x => x.trades.errors.has('trade-insert-error'), html<TradeInsertErrorFeedback>`...`)}
     * ```
     *
     * @example
     * Errors can be bound to collectively:
     * ```ts
     * <pre>${x => x.trades.errors.messages}</pre>
     * ```
     *
     * @public
     */
    readonly errors: ErrorMap<EventDetailMap>;
    /**
     * An api to allow the observation of values and arbitrary bindings outside the template engine.
     *
     * @example
     * If you bind with a lookup token, TS attains the value type information from the store's interface.
     * ```ts
     * this.store.binding(x => x.ready, value => {...}); // value is a boolean
     * this.store.binding(x => x.someNumericProp, value => {...}); // value is a number
     * ```
     *
     * @example
     * If you use a string token, then you need to provide the value type information.
     * ```ts
     * this.store.binding<boolean>('ready', value => {...}); // value is a boolean
     * this.store.binding<number>('someNumericProp', value => {...}); // value is a number
     * ```
     *
     * @example
     * You can create your own derived bindings that are not already pre-defined as getters on the store's interface.
     * ```ts
     * this.store.binding(x => x.prop1 * x.prop2, value => {...});
     * ```
     *
     * @example
     * Such bindings can even be volatile providing you pass true for isVolatileBinding.
     * ```ts
     * this.store.binding(x => x.someToggle ? x.prop1 : x.prop2, value => {...}, true);
     * ```
     *
     * @example
     * You can also use the underlying BindingObserver api should you prefer.
     * ```ts
     * this.store.binding(x => x.someData).subscribe({
     *   handleChange(binding, args) {
     *     ...
     *   }
     * })
     * ```
     *
     * @param token - A store lookup token which can take various forms.
     * @param subscriberChangeCallback - {@link SubscriberChangeCallback}
     * @param isVolatileBinding - Indicates the binding is volatile.
     * @param context - {@link @microsoft/fast-element#ExecutionContext}
     * @returns An rxjs Observable
     *
     * @public
     * @privateRemarks
     * We might want to break this out into overloads like bindingAsRx.
     */
    binding<TReturn = any, TParent = any>(token: ((store: this) => TReturn) | keyof this, subscriberChangeCallback?: SubscriberChangeCallback<TReturn> | undefined, isVolatileBinding?: boolean, context?: ExecutionContext): BindingObserver<this, TReturn, TParent>;
    /**
     * An api to allow the observation of values and arbitrary bindings as Rx observables.
     *
     * @example
     * ```ts
     * const entireStore$ = this.store.bindingAsRx().subscribe(value => {...});
     * ```
     *
     * @returns An rxjs Observable of the entire store.
     *
     * @public
     * @privateRemarks
     * This needs to be reviewed, as during initial testing rxjs based bindings seemed to only tick once.
     */
    bindingAsRx(): RXObservable<this>;
    /**
     * {@inheritDoc Store.bindingAsRx}
     *
     * @example
     * ```ts
     * const ready$ = this.store.bindingAsRx('ready').subscribe(value => {...});
     * ```
     *
     * @param key - Store property key.
     * @returns An rxjs Observable of that key's value.
     */
    bindingAsRx<TKey extends keyof this>(key: TKey): RXObservable<this[TKey]>;
    /**
     * {@inheritDoc Store.bindingAsRx}
     *
     * @example
     * ```ts
     * const prop1$ = this.store.bindingAsRx(x => x.prop1).subscribe(value => {...});
     * ```
     *
     * @param getter - A function that returns a value from the store.
     * @returns An rxjs Observable of the value.
     */
    bindingAsRx<TReturn>(getter: (store: this) => TReturn): RXObservable<TReturn>;
    /**
     * {@inheritDoc Store.bindingAsRx}
     *
     * @param token - Store property key or a function that returns a value from the store.
     * @returns An rxjs Observable of the value.
     */
    bindingAsRx<TReturn>(token?: ((store: this) => TReturn) | keyof this): RXObservable<TReturn | this>;
}
/**
 * Root store interface.
 *
 * @public
 * @remarks
 * The public interface available on the injected store fragment.
 */
export interface StoreRoot extends Store {
    /**
     * The store root element.
     *
     * @public
     */
    readonly element: HTMLElement;
    /**
     * The ready status of the store root.
     * @returns boolean
     *
     * @public
     */
    readonly ready: boolean;
    onConnected(event: CustomEvent<HTMLElement>): void;
    onDisconnected(event: CustomEvent<void>): void;
    onReady(event: CustomEvent<boolean>): void;
}
/**
 * Store connectable interface.
 * @typeParam TStoreRoot - The store root interface.
 *
 * @public
 */
export interface StoreConnectable<TStoreRoot extends StoreRoot> {
    /**
     * Connects this store fragment.
     * @param root - The store root fragment.
     *
     * @public
     */
    connect(root: TStoreRoot): void;
    /**
     * Disconnects this store fragment.
     * @param root - The store root fragment.
     *
     * @public
     */
    disconnect(root: TStoreRoot): void;
}
/**
 * @typeParam TStore - The store interface.
 *
 * @public
 */
export type SubscriberChangeHandler<TStore extends Store, TReturn = any> = (binding: (x?: TStore, context?: ExecutionContext) => TReturn, args: BindingObserver) => void;
/**
 * @typeParam TStore - The store interface.
 * @typeParam TReturn - The return type.
 *
 * @public
 */
export interface StoreSubscriber<TStore extends Store, TReturn = any> {
    handleChange: SubscriberChangeHandler<TStore, TReturn>;
}
/**
 * @typeParam TStore - The store interface.
 *
 * @public
 */
export type StoreBinding<TStore extends Store> = (store: TStore) => any;
/**
 * Creates a dependency injection key for the store being registered.
 * @param Base - The store fragment class.
 * @param name - The name of your store fragment, mostly used for internal logging.
 * @returns The created key.
 *
 * @example
 * ```ts
 * export const TradeEntry = registerStore(DefaultTradeEntry, 'TradeEntry');
 * ```
 *
 * @public
 */
export declare const registerStore: <K extends Key>(Base: Constructable<Store>, name?: string) => import("@microsoft/fast-foundation").InterfaceSymbol<K>;
/**
 * The abstract store that concrete store fragments must extend, which differs from the {@link AbstractStoreRoot}.
 * @typeParam TStore - The store interface.
 * @typeParam TEventDetailMap - The store's public event detail map.
 * @typeParam TInternalEventDetailMap - The store's internal event detail map.
 * @typeParam TStoreRoot - The store root interface.
 *
 * @public
 */
export declare abstract class AbstractStore<TStore extends Store, TEventDetailMap extends EventDetailMap, TInternalEventDetailMap extends EventDetailMap = TEventDetailMap, TStoreRoot extends StoreRoot = StoreRoot> implements Store {
    /**
     * The store fragment's event listener map.
     *
     * @internal
     */
    private eventListenerMap;
    /**
     * The store root fragment.
     *
     * @public
     */
    protected root: TStoreRoot;
    /**
     * The child store fragments this store fragment manages.
     *
     * @internal
     */
    protected readonly storeFragments: Store[];
    /**
     * @param storeFragments - The child store fragments this store fragment will manage.
     */
    constructor(...storeFragments: Store[]);
    /**
     * The value commit proxy handler.
     *
     * @internal
     * @privateRemarks
     * This provides a way to do things on value setting like logging, validation, reject or tweak values, have a
     * middleware type setup to perhaps add dev tool state time travel functionality etc. Create a hash of current store
     * state and push to history for example. If the dev fails to use this commit proxy, ie. this.commit.prop = value,
     * it's not the end of the world as value will still be set. this.commit is typed the same as the main interface.
     * Commit mutations should always be synchronous.
     */
    protected readonly commitHandler: ProxyHandler<this>;
    /**
     * The value commit proxy.
     *
     * @example
     * ```ts
     * this.commit.propX = value;
     * ```
     *
     * @beta
     * @remarks
     * this.commit has the same interface as the store itself, so props are strongly typed.
     */
    protected readonly commit: this;
    /**
     * Alternative value commit api.
     * @param key - The property key from the store's interface.
     * @param value - The value to set.
     *
     * @beta
     * @privateRemarks
     * We should have only one way to do things, so we can ga gather user feedback and keep or kill this.
     */
    protected commitValue<K extends keyof TStore>(key: K, value: TStore[K]): void;
    /** {@inheritDoc Store.binding} */
    binding<TReturn = any, TParent = any>(token: ((store: this) => TReturn) | keyof this, subscriberChangeCallback?: SubscriberChangeCallback<TReturn> | undefined, isVolatileBinding?: boolean, context?: ExecutionContext): BindingObserver<this, TReturn, TParent>;
    /** {@inheritDoc Store.bindingAsRx} */
    bindingAsRx(): RXObservable<this>;
    /** {@inheritDoc Store.bindingAsRx} */
    bindingAsRx<TKey extends keyof this>(key: TKey): RXObservable<this[TKey]>;
    /** {@inheritDoc Store.bindingAsRx} */
    bindingAsRx<TReturn>(getter: (store: this) => TReturn): RXObservable<TReturn>;
    /** {@inheritDoc StoreConnectable.connect} */
    connect(root: TStoreRoot): void;
    /** {@inheritDoc StoreConnectable.disconnect} */
    disconnect(root: TStoreRoot): void;
    /** {@inheritDoc Store.errors} */
    errors: ErrorMap<TEventDetailMap & TInternalEventDetailMap>;
    /** {@inheritDoc Store.name} */
    get name(): string;
    /**
     * Connects the store fragment's child store fragments if any.
     * @param root - The store root fragment.
     *
     * @internal
     * @privateRemarks
     * We have limited the connect and disconnect apis to store root only, so we need to cast these to invoke the same.
     */
    protected connectStoreFragments(root: TStoreRoot): void;
    /**
     * Disconnects the store fragment's child store fragments if any.
     * @param root - The store root fragment.
     *
     * @internal
     * @privateRemarks
     * We have limited the connect and disconnect apis to store root only, so we need to cast these to invoke the same.
     */
    protected disconnectStoreFragments(root: TStoreRoot): void;
    /**
     * Logs events.
     * @param key - The event key from the store fragment's event detail map.
     * @param isAsyncHandler - A flag to indicate if the handler is async.
     *
     * @internal
     */
    private logEvent;
    /**
     * Sets event listeners on the store fragment's internal map.
     * @param key - The event key from the store fragment's event detail map.
     * @param listener - The event listener.
     * @returns The event listener.
     *
     * @internal
     */
    private mappedListener;
    /**
     * Creates an event listener.
     * @typeParam TDetail - The CustomEvent detail.
     * @param key - The event key from the store fragment's event detail map.
     * @param token - The function handling the event.
     * @returns The event listener.
     *
     * @public
     * @remarks
     * You can think of this like a `reducer` in the redux sense. You are allowed to commit values to the store in these
     * synchronous handlers.
     */
    protected createListener: <TDetail = void>(key: keyof (TEventDetailMap & TInternalEventDetailMap), token: (detail: TDetail) => void) => EventListener;
    /**
     * Creates an async event listener.
     * @typeParam TDetail - The CustomEvent detail.
     * @param key - The event key from the store fragment's event detail map.
     * @param token - The async function handling the event.
     * @returns The event listener.
     *
     * @public
     * @remarks
     * You can think of this like an `effect` in the redux sense. You should not commit values to the store in these,
     * instead raise subsequent events to be handled synchronously, where commits are allowed.
     */
    protected createAsyncListener: <TDetail = void, TReturn = void>(key: keyof (TEventDetailMap & TInternalEventDetailMap), token: (detail: TDetail) => Promise<TReturn>) => EventListener;
    /**
     * Creates an error event listener.
     * @typeParam TDetail - The CustomEvent detail.
     * @param key - The event key from the store fragment's event detail map.
     * @param token - The function handling the event.
     * @returns The event listener.
     *
     * @public
     * @remarks
     * This logs and stores errors by event key in the store fragment's {@link ErrorMap}, allowing multiple errors to
     * co-exist, and be presentable to the user via the UI for further action or dismissal.
     */
    protected createErrorListener: <TDetail extends Error = Error>(key: keyof (TEventDetailMap & TInternalEventDetailMap), token?: (detail: TDetail) => void) => EventListener;
    /**
     * Add the store fragment's event listeners to the root element.
     * @param element - The store root element.
     *
     * @internal
     * @privateRemarks
     * Called via connect() when the root element changes.
     */
    protected addEventListeners(element: HTMLElement): void;
    /**
     * Removes the store fragment's event listeners from the root element.
     * @param element - The store root element.
     *
     * @internal
     * @privateRemarks
     * Called via disconnect() when the root element changes.
     */
    protected removeEventListeners(element: HTMLElement): void;
    /**
     * Emit events to the stores directly via the standard event flow.
     * @param key - The event key from the store fragment's event detail map.
     * @param detail - The event detail object to send with the event.
     * @param options - The event options. By default bubbles and composed.
     *
     * @public
     * @remarks
     * By default we allow stores to emit everything in their TEventDetailMap, however TInternalEventDetailMap can be
     * provided to define additional internal events like x-success, x-error etc, or public events from other store
     * fragments.
     */
    protected emit<K extends keyof (TEventDetailMap & TInternalEventDetailMap) & string>(...args: (TEventDetailMap & TInternalEventDetailMap)[K] extends void ? [key: K] : [key: K, detail: (TEventDetailMap & TInternalEventDetailMap)[K]]): void;
}
/**
 * Store root event key to event detail map.
 * @public
 */
export type StoreRootEventDetailMap = {
    'store-connected': HTMLElement;
    'store-disconnected': void;
    'store-ready': boolean;
};
/**
 * The abstract store root that concrete store roots must extend.
 * @typeParam TStore - The store interface.
 * @typeParam TEventDetailMap - The store's public event detail map.
 * @typeParam TInternalEventDetailMap - The store's internal event detail map.
 *
 * @public
 */
export declare abstract class AbstractStoreRoot<TStore extends StoreRoot, TEventDetailMap extends StoreRootEventDetailMap = StoreRootEventDetailMap, TInternalEventDetailMap extends EventDetailMap = TEventDetailMap> extends AbstractStore<TStore, TEventDetailMap, TInternalEventDetailMap> implements StoreRoot {
    /** {@inheritDoc StoreRoot.ready} */
    ready: boolean;
    /** {@inheritDoc StoreRoot.element} */
    element: HTMLElement;
    /**
     * @public
     * @privateRemarks
     * Set to self
     */
    protected root: this;
    /**
     * Internal change handler for the element property.
     * @param prev - The previous element value.
     * @param next - The next element value.
     *
     * @internal
     * @privateRemarks
     * Manages the store listeners on change.
     */
    private elementChanged;
    onConnected: EventListener;
    onDisconnected: EventListener;
    onReady: EventListener;
}
//# sourceMappingURL=foundationStore.d.ts.map