import { __awaiter, __decorate } from "tslib";
import { isDev } from '@genesislcap/foundation-utils';
import { defaultExecutionContext, observable, Observable as FASTObservable, } from '@microsoft/fast-element';
import { DI } from '@microsoft/fast-foundation';
import safeStringify from 'fast-safe-stringify';
import { of } from 'rxjs';
import { distinctUntilChanged, map } from 'rxjs/operators';
import { logger } from '../utils';
import { createErrorMap } from './errorMap';
const getBindArgs = (...args) => {
    const [token] = args;
    let binding = (x) => x;
    if (typeof token === 'function') {
        binding = token;
    }
    else if (typeof token === 'string') {
        binding = (x) => x[token];
    }
    return binding;
};
/**
 * Creates a dependency injection key for the store being registered.
 * @param Base - The store fragment class.
 * @param name - The name of your store fragment, mostly used for internal logging.
 * @returns The created key.
 *
 * @example
 * ```ts
 * export const TradeEntry = registerStore(DefaultTradeEntry, 'TradeEntry');
 * ```
 *
 * @public
 */
export const registerStore = (Base, name = 'store') => {
    class Store extends Base {
        get name() {
            return name;
        }
    }
    return DI.createInterface((x) => x.singleton(Store));
};
const logValue = (value) => safeStringify(value, undefined, 2, {
    depthLimit: 1,
    edgesLimit: 1,
});
const internalEmitOptions = {
    bubbles: true,
    composed: true,
    cancelable: true,
};
/**
 * The abstract store that concrete store fragments must extend, which differs from the {@link AbstractStoreRoot}.
 * @typeParam TStore - The store interface.
 * @typeParam TEventDetailMap - The store's public event detail map.
 * @typeParam TInternalEventDetailMap - The store's internal event detail map.
 * @typeParam TStoreRoot - The store root interface.
 *
 * @public
 */
export class AbstractStore {
    /**
     * @param storeFragments - The child store fragments this store fragment will manage.
     */
    constructor(...storeFragments) {
        /**
         * The store fragment's event listener map.
         *
         * @internal
         */
        this.eventListenerMap = new Map();
        /**
         * The child store fragments this store fragment manages.
         *
         * @internal
         */
        this.storeFragments = [];
        /**
         * The value commit proxy handler.
         *
         * @internal
         * @privateRemarks
         * This provides a way to do things on value setting like logging, validation, reject or tweak values, have a
         * middleware type setup to perhaps add dev tool state time travel functionality etc. Create a hash of current store
         * state and push to history for example. If the dev fails to use this commit proxy, ie. this.commit.prop = value,
         * it's not the end of the world as value will still be set. this.commit is typed the same as the main interface.
         * Commit mutations should always be synchronous.
         */
        this.commitHandler = {
            set: (target, prop, next) => {
                const prev = target[prop];
                if (next === prev) {
                    return true;
                }
                if (isDev()) {
                    logger.debug(`\n'${this.name}':\n   Committed ${prop} = ${logValue(next)}.\n   Previous ${prop} = ${logValue(prev)}`);
                }
                target[prop] = next;
                return true;
            },
        };
        /**
         * The value commit proxy.
         *
         * @example
         * ```ts
         * this.commit.propX = value;
         * ```
         *
         * @beta
         * @remarks
         * this.commit has the same interface as the store itself, so props are strongly typed.
         */
        this.commit = new Proxy(this, this.commitHandler);
        /** {@inheritDoc Store.errors} */
        this.errors = createErrorMap(logger.error);
        /**
         * Creates an event listener.
         * @typeParam TDetail - The CustomEvent detail.
         * @param key - The event key from the store fragment's event detail map.
         * @param token - The function handling the event.
         * @returns The event listener.
         *
         * @public
         * @remarks
         * You can think of this like a `reducer` in the redux sense. You are allowed to commit values to the store in these
         * synchronous handlers.
         */
        this.createListener = (key, token) => {
            const listener = ({ detail }) => {
                this.logEvent(key);
                return token(detail);
            };
            return this.mappedListener(key, listener);
        };
        /**
         * Creates an async event listener.
         * @typeParam TDetail - The CustomEvent detail.
         * @param key - The event key from the store fragment's event detail map.
         * @param token - The async function handling the event.
         * @returns The event listener.
         *
         * @public
         * @remarks
         * You can think of this like an `effect` in the redux sense. You should not commit values to the store in these,
         * instead raise subsequent events to be handled synchronously, where commits are allowed.
         */
        this.createAsyncListener = (key, token) => {
            const listener = ({ detail }) => __awaiter(this, void 0, void 0, function* () {
                this.logEvent(key, true);
                return token(detail);
            });
            return this.mappedListener(key, listener);
        };
        /**
         * Creates an error event listener.
         * @typeParam TDetail - The CustomEvent detail.
         * @param key - The event key from the store fragment's event detail map.
         * @param token - The function handling the event.
         * @returns The event listener.
         *
         * @public
         * @remarks
         * This logs and stores errors by event key in the store fragment's {@link ErrorMap}, allowing multiple errors to
         * co-exist, and be presentable to the user via the UI for further action or dismissal.
         */
        this.createErrorListener = (key, token) => {
            const listener = ({ detail }) => {
                this.logEvent(key);
                this.errors.set(key, detail);
                return token && token(detail);
            };
            return this.mappedListener(key, listener);
        };
        this.storeFragments = storeFragments;
    }
    /**
     * Alternative value commit api.
     * @param key - The property key from the store's interface.
     * @param value - The value to set.
     *
     * @beta
     * @privateRemarks
     * We should have only one way to do things, so we can ga gather user feedback and keep or kill this.
     */
    commitValue(key, value) {
        this.commit[key] = value;
    }
    /** {@inheritDoc Store.binding} */
    binding(token, subscriberChangeCallback, isVolatileBinding = false, context = defaultExecutionContext) {
        const valueGetter = getBindArgs(token);
        const initialSubscriber = {
            handleChange: (binding, args) => {
                if (subscriberChangeCallback) {
                    subscriberChangeCallback(binding(this, context), args);
                }
            },
        };
        const bindingObserver = FASTObservable.binding(valueGetter, initialSubscriber, isVolatileBinding);
        bindingObserver.observe(this, context);
        return bindingObserver;
    }
    /** {@inheritDoc Store.bindingAsRx} */
    bindingAsRx(token) {
        const valueGetter = getBindArgs(token);
        return of(this).pipe(map((store) => valueGetter(store)), distinctUntilChanged());
    }
    /** {@inheritDoc StoreConnectable.connect} */
    connect(root) {
        this.root = root;
        this.addEventListeners(root.element);
        this.connectStoreFragments(root);
    }
    /** {@inheritDoc StoreConnectable.disconnect} */
    disconnect(root) {
        this.removeEventListeners(root.element);
        this.disconnectStoreFragments(root);
        this.root = undefined;
    }
    /** {@inheritDoc Store.name} */
    get name() {
        return 'store';
    }
    /**
     * Connects the store fragment's child store fragments if any.
     * @param root - The store root fragment.
     *
     * @internal
     * @privateRemarks
     * We have limited the connect and disconnect apis to store root only, so we need to cast these to invoke the same.
     */
    connectStoreFragments(root) {
        this.storeFragments.forEach((store) => {
            const connectableStore = store;
            connectableStore.connect(root);
        });
    }
    /**
     * Disconnects the store fragment's child store fragments if any.
     * @param root - The store root fragment.
     *
     * @internal
     * @privateRemarks
     * We have limited the connect and disconnect apis to store root only, so we need to cast these to invoke the same.
     */
    disconnectStoreFragments(root) {
        this.storeFragments.forEach((store) => {
            const connectableStore = store;
            connectableStore.disconnect(root);
        });
    }
    /**
     * Logs events.
     * @param key - The event key from the store fragment's event detail map.
     * @param isAsyncHandler - A flag to indicate if the handler is async.
     *
     * @internal
     */
    logEvent(key, isAsyncHandler) {
        if (isDev()) {
            logger.info(`'${this.name}': Handle '${String(key)}' store event ${isAsyncHandler ? '[async]' : ''}`);
        }
    }
    /**
     * Sets event listeners on the store fragment's internal map.
     * @param key - The event key from the store fragment's event detail map.
     * @param listener - The event listener.
     * @returns The event listener.
     *
     * @internal
     */
    mappedListener(key, listener) {
        this.eventListenerMap.set(key, listener);
        return listener;
    }
    /**
     * Add the store fragment's event listeners to the root element.
     * @param element - The store root element.
     *
     * @internal
     * @privateRemarks
     * Called via connect() when the root element changes.
     */
    addEventListeners(element) {
        for (const [key, listener] of this.eventListenerMap) {
            element.addEventListener(key, listener);
        }
    }
    /**
     * Removes the store fragment's event listeners from the root element.
     * @param element - The store root element.
     *
     * @internal
     * @privateRemarks
     * Called via disconnect() when the root element changes.
     */
    removeEventListeners(element) {
        for (const [key, listener] of this.eventListenerMap) {
            element.removeEventListener(key, listener);
        }
    }
    /**
     * Emit events to the stores directly via the standard event flow.
     * @param key - The event key from the store fragment's event detail map.
     * @param detail - The event detail object to send with the event.
     * @param options - The event options. By default bubbles and composed.
     *
     * @public
     * @remarks
     * By default we allow stores to emit everything in their TEventDetailMap, however TInternalEventDetailMap can be
     * provided to define additional internal events like x-success, x-error etc, or public events from other store
     * fragments.
     */
    emit(...args) {
        const [key, detail] = args;
        const hasDetail = detail !== undefined;
        if (isDev()) {
            logger.info(`'${this.name}': Emit '${key}' store event${hasDetail ? ` with ${logValue(detail)}` : ''}`);
        }
        this.root.element.dispatchEvent(new CustomEvent(key, Object.assign({ detail }, internalEmitOptions)));
    }
}
/**
 * The abstract store root that concrete store roots must extend.
 * @typeParam TStore - The store interface.
 * @typeParam TEventDetailMap - The store's public event detail map.
 * @typeParam TInternalEventDetailMap - The store's internal event detail map.
 *
 * @public
 */
export class AbstractStoreRoot extends AbstractStore {
    constructor() {
        super(...arguments);
        /** {@inheritDoc StoreRoot.ready} */
        this.ready = false;
        /**
         * @public
         * @privateRemarks
         * Set to self
         */
        this.root = this;
        this.onConnected = this.createListener('store-connected', (element) => (this.commit.element = element));
        this.onDisconnected = this.createListener('store-disconnected', () => (this.commit.element = undefined));
        this.onReady = this.createListener('store-ready', (ready) => (this.commit.ready = ready));
    }
    /**
     * Internal change handler for the element property.
     * @param prev - The previous element value.
     * @param next - The next element value.
     *
     * @internal
     * @privateRemarks
     * Manages the store listeners on change.
     */
    elementChanged(prev, next) {
        if (prev) {
            this.removeEventListeners(prev);
            this.disconnectStoreFragments(this);
        }
        if (next) {
            this.addEventListeners(next);
            this.connectStoreFragments(this);
        }
    }
}
__decorate([
    observable
], AbstractStoreRoot.prototype, "ready", void 0);
__decorate([
    observable
], AbstractStoreRoot.prototype, "element", void 0);
